/*
   Name:    testOOo2.rex
   Purpose: test whether OpenOffice.org/StarOffice can be automated via ooRexx,
            demonstrate some of the reflection methods for finding out informaton
            about UNO_PROXY objects at runtime
   Author:  Rony G. Flatscher
   Date:    2006-01-03
   Needs:   UNO.CLS (Universal Network Objects support, component technology on
                     which OOo/SO are built)

   Remarks: For further information on programming OOo/SO and getting at ooRexx
            nutshell examples that jump-start you please refer to
            <http://wi.wu-wien.ac.at/rgf/diplomarbeiten/index.html#bakk_07> (Andreas Ahammer);

            Also a must, the homepage and developer information (including the *excellent*
            Developer guide for OOo 2.0 dated 2005-05 or later): <http://www.OpenOffice.org>
*/

say "testOOo.rex: started to run; please be patient on the very first invocation"
say "             as creating the full Java support for OpenOffice.org (OOo) or"
say "             StarOffice (SO) takes a while initially."
say
   /* create a desktop object, request the XDesktop interface from it and
      request the XComponentLoader interface from that                        */
say "... getting at a componentLoader ..."
componentLoader = UNO.createDesktop()~XDesktop~XComponentLoader
say "    componentLoader~uno.getTypeName  ="pp(componentLoader~uno.getTypeName)
say "    componentLoader~uno.getDefinition="ppd(componentLoader~uno.getDefinition)
call eyeCatcher

   /* use the XComponentLoader interface function "loadComponentFromUrl()" to
      create a writer component (= word processor document)                   */
say "... loading a writerComponent ..."
writerComponent = componentLoader~loadComponentFromURL("private:factory/swriter", "_blank", 0, .UNO~noProps)
say "    writerComponent~uno.getTypeName  ="pp(writerComponent~uno.getTypeName)
say "    writerComponent~uno.getDefinition="ppd(writerComponent~uno.getDefinition)
call eyeCatcher

say "    writerComponent~uno.findInterfaceWithMember('getText')="pp(writerComponent~uno.findInterfaceWithMember('getText'))
call eyeCatcher
   /* get the XTextDocument interface from the writerComponent and use its function
      getText() to get a hold of the text object                              */
say "... getting the text object ..."
text = writerComponent~XTextDocument~getText

   /* set the text of the text object to salute the world of ooRexx           */
say "... setting the text object ..."
say "    text~uno.getTypeName  ="pp(text~uno.getTypeName)
say "    text~uno.getDefinition="ppd(text~uno.getDefinition)
text~setString("Hello OpenOffice.org/StarOffice, this is ooRexx speaking! ("date("S") time()")")
call eyeCatcher

say
say "done."
say
say "(You may want to re-run this script to see the startup speed improvement!)"

::requires UNO.CLS   /* directive to get the UNO support for ooRexx, defines utility classes, routines */

::routine eyeCatcher /* output eyecatcher lines */

   say "   " copies("---------   ", 6)
   say "   " copies("\\\---///   ", 6)
   say "   " copies("---------   ", 6)
   say
